/*
 * Copyright (C) 2008 KenD00
 * 
 * This file is part of ConvTableView.
 * 
 * ConvTableView is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package convtableview.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;

import javax.swing.ButtonGroup;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.JTree;
import javax.swing.ListSelectionModel;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.table.TableModel;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

import dumphd.bdplus.ConversionTable;
import dumphd.util.FileSource;
import dumphd.util.MessagePrinter;
import dumphd.util.Utils;


/**
 * @author KenD00
 *
 */
public class ConvTableView {

   private final static String PROGRAM_NAME = "ConvTableView";
   private final static String PROGRAM_AUTHOR = "KenD00";
   private final static String PROGRAM_VERSION = "0.3";

   private final static UIManager.LookAndFeelInfo laf[] = UIManager.getInstalledLookAndFeels();

   private MessagePrinter out = null;

   private JFrame mainFrame = null;
   private JSplitPane horizontalSplitPane = null;
   private JSplitPane verticalSplitPane = null;

   private JFileChooser sourceChooser = null;

   private JTree tableTree = null;
   private DefaultMutableTreeNode tableTreeRootNode = null;
   private JTable contentTable = null;
   private TableModel contentTableEmptyRecord = null;
   private JScrollPane contentTableScroller = null;
   private JTextArea logArea = null;


   public ConvTableView() {
      JMenuBar menuBar = null;
      JMenu menu = null;
      JMenuItem menuItem = null;
      ButtonGroup buttonGroup = null;
      JRadioButtonMenuItem radioButtonMenuItem = null;

      String systemLafClassName = UIManager.getSystemLookAndFeelClassName();

      Font monoFont = Font.decode(Font.MONOSPACED);

      // Set the Look & Feel to the System Look & Feel
      try {
         UIManager.setLookAndFeel(systemLafClassName);
      }
      catch (UnsupportedLookAndFeelException e1) {
         // TODO: What to do?
      }
      catch (Exception e2) {
         // TODO: What to do?
      }

      // Set up the main JFrame
      mainFrame = new JFrame(ConvTableView.PROGRAM_NAME + " " + ConvTableView.PROGRAM_VERSION);
      mainFrame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
      mainFrame.addWindowListener(new WindowAdapter() {
         public void windowClosing(WindowEvent e) {
            close();
         }
      });

      // ######################
      // ###****************###
      // ###*** Menu bar ***###
      // ###****************###
      // ######################
      menuBar = new JMenuBar();
      // *****************
      // *** File menu ***
      // *****************
      menu = new JMenu("File");
      menuItem = new JMenuItem("Open");
      menuItem.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            int returnVal = sourceChooser.showOpenDialog(mainFrame);
            if (returnVal == JFileChooser.APPROVE_OPTION) {
               //TODO: Read in another thread to stay responsive 
               ((DefaultTreeModel)(tableTree.getModel())).setRoot(tableTreeRootNode);
               //tableTreeRootNode.removeAllChildren();
               contentTable.setModel(contentTableEmptyRecord);
               File selectedFile = sourceChooser.getSelectedFile().getAbsoluteFile();
               try {
                  FileSource tableSource = new FileSource(selectedFile, FileSource.R_MODE);
                  ConversionTable conversionTable = new ConversionTable(tableSource);
                  if (tableSource.getPosition() == tableSource.size()) {
                     out.println(String.format("%1$s successfully loaded", tableSource.getFile().getAbsolutePath()));
                  } else {
                     out.println(String.format("%1%s: Warning, data behind table end", tableSource.getFile().getAbsolutePath()));
                  }
                  tableSource.close();
                  ConversionTableWrapper wrapper = new ConversionTableWrapper(conversionTable, tableSource.getFile().getAbsolutePath());
                  ((DefaultTreeModel)(tableTree.getModel())).setRoot(wrapper);
                  //tableTreeRootNode.add(wrapper);
               }
               catch (IOException ex) {
                  out.println(ex.getMessage());
               }
               //((DefaultTreeModel)(tableTree.getModel())).reload();
            }
         }
      });
      menu.add(menuItem);
      menu.addSeparator();
      menuItem = new JMenuItem("Exit");
      menuItem.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            close();
         }
      });
      menu.add(menuItem);
      menuBar.add(menu);
      // ************************
      // *** Look & Feel menu ***
      // ************************
      menu = new JMenu("Look & Feel");
      LafActionListener lal = new LafActionListener();
      buttonGroup = new ButtonGroup();
      for (int i = 0; i < ConvTableView.laf.length; i++) {
         String lafClassName = laf[i].getClassName();
         if (lafClassName.equals(systemLafClassName)) {
            radioButtonMenuItem = new JRadioButtonMenuItem(laf[i].getName(), true);
         }
         else {
            radioButtonMenuItem = new JRadioButtonMenuItem((laf[i].getName()), false);
         }
         radioButtonMenuItem.setActionCommand(lafClassName);
         radioButtonMenuItem.addActionListener(lal);
         buttonGroup.add(radioButtonMenuItem);
         menu.add(radioButtonMenuItem);
      }
      menuBar.add(menu);
      // *****************
      // *** Help menu ***
      // *****************
      menu = new JMenu("Help");
      //menuItem = new JMenuItem("Help");
      //menu.add(menuItem);
      //menu.addSeparator();
      menuItem = new JMenuItem("About");
      menuItem.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            JOptionPane.showMessageDialog (mainFrame, ConvTableView.PROGRAM_NAME + " " + ConvTableView.PROGRAM_VERSION + " by " + ConvTableView.PROGRAM_AUTHOR +
                  "\n\nA small utility to show the contents of a BD+ Conversion Table.\n",
                  "About", JOptionPane.INFORMATION_MESSAGE);
         }
      });
      menu.add(menuItem);
      menuBar.add(menu);
      mainFrame.setJMenuBar(menuBar);

      // ### Setting up the FileChooser ###
      sourceChooser = new JFileChooser();
      sourceChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
      sourceChooser.setMultiSelectionEnabled(false);


      // #######################
      // ###*****************###
      // ###*** Tree view ***###
      // ###*****************###
      // #######################
      JPanel treePanel = new JPanel(new BorderLayout());
      treePanel.setOpaque(true);
      tableTreeRootNode = new DefaultMutableTreeNode("ConvTableView");
      DefaultTreeModel dm = new DefaultTreeModel(tableTreeRootNode, true);
      tableTree = new JTree(dm);
      //tableTree.setRootVisible(false);
      if (monoFont != null) {
         tableTree.setFont(monoFont);
      }
      tableTree.getSelectionModel().addTreeSelectionListener(new TreeSelectionListener() {
         public void valueChanged(TreeSelectionEvent e) {
            TreePath newPath = e.getNewLeadSelectionPath();
            if (newPath != null) {
               Object lastComponent = newPath.getLastPathComponent();
               if (lastComponent instanceof SegmentWrapper) {
                  contentTableScroller.getViewport().setViewPosition(new Point(0, 0));
                  contentTable.setModel((SegmentWrapper)lastComponent);
               } else {
                  contentTableScroller.getViewport().setViewPosition(new Point(0, 0));
                  contentTable.setModel(contentTableEmptyRecord);
               }
            }
         }
      });
      treePanel.add(new JScrollPane(tableTree, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED), BorderLayout.CENTER);


      // ###########################
      // ###*********************###
      // ###*** Content Table ***###
      // ###*********************###
      // ###########################
      JPanel contentPanel = new JPanel(new BorderLayout());
      contentPanel.setOpaque(true);
      contentTableEmptyRecord = new SegmentWrapper(null, null);
      contentTable = new JTable(contentTableEmptyRecord);
      if (monoFont != null) {
         contentTable.setFont(monoFont);
      }
      contentTable.setAutoCreateColumnsFromModel(false);
      contentTable.setAutoResizeMode(JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS);
      contentTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      contentTable.setShowGrid(true);
      contentTable.getTableHeader().setReorderingAllowed(false);
      contentTableScroller = new JScrollPane(contentTable, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
      contentPanel.add(contentTableScroller, BorderLayout.CENTER);


      // #################
      // ###***********###
      // ###*** Log ***###
      // ###***********###
      // #################
      JPanel logPanel = new JPanel(new BorderLayout());
      logPanel.setOpaque(true);
      // ****************
      // *** Menu bar ***
      // ****************
      menuBar = new JMenuBar();
      menu = new JMenu("Log");
      menuItem = new JMenuItem("Clear");
      menuItem.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            Document logDocument = logArea.getDocument();
            try {
               logDocument.remove(0, logDocument.getLength());
            }
            catch (BadLocationException ex) {
               // TODO: What to do?
            }
         }
      });
      menu.add(menuItem);
      menuBar.add(menu);
      logPanel.add(menuBar, BorderLayout.PAGE_START);
      // *********************
      // *** Log text area ***
      // *********************
      logArea = new JTextArea();
      if (monoFont != null) {
         logArea.setFont(monoFont);
      }
      logArea.setEditable(false);
      JScrollPane scrollPane = new JScrollPane(logArea, ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
      scrollPane.setMinimumSize(new Dimension(64, 64));
      logPanel.add(scrollPane, BorderLayout.CENTER);
      out = new TextAreaMessagePrinter();
      Utils.setMessagePrinter(out);


      // ##############################
      // ###************************###
      // ###*** Mainframe layout ***###
      // ###************************###
      // ##############################
      // ******************
      // *** Splitpanes ***
      // ******************
      horizontalSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, true);
      horizontalSplitPane.setResizeWeight(0.5);
      verticalSplitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT, true);
      verticalSplitPane.setResizeWeight(0.25);
      horizontalSplitPane.setLeftComponent(treePanel);
      horizontalSplitPane.setRightComponent(contentPanel);
      verticalSplitPane.setTopComponent(horizontalSplitPane);
      verticalSplitPane.setBottomComponent(logPanel);
      // ************************
      // *** Mainframe layout ***
      // ************************
      mainFrame.setContentPane(verticalSplitPane);
      //mainFrame.pack();
      mainFrame.setSize(800, 600);
   }

   private void close() {
      mainFrame.dispose();
      System.exit(0);
   }

   public void show() {
      mainFrame.setVisible(true);
   }

   public void dispose() {
      mainFrame.dispose();
   }


   public static void main(String[] args) {
      ConvTableView tableView = new ConvTableView();
      tableView.show();
   }


   private class LafActionListener implements ActionListener {
      public void actionPerformed(ActionEvent e) {
         String lafClassName = e.getActionCommand();
         try {
            UIManager.setLookAndFeel(lafClassName);
            SwingUtilities.updateComponentTreeUI(mainFrame);
            SwingUtilities.updateComponentTreeUI(sourceChooser);
            logArea.setCaretPosition(logArea.getDocument().getLength());
            horizontalSplitPane.resetToPreferredSizes();
            verticalSplitPane.resetToPreferredSizes();
         }
         catch (UnsupportedLookAndFeelException e1) {
            // TODO: What to do?
         }
         catch (Exception e2) {
            // TODO: What to do?
         }
      }
   }

   private class TextAreaMessagePrinter implements MessagePrinter {

      public void print(final String msg) {
         Runnable runMe = new Runnable() {
            public void run() {
               logArea.append(msg);
            }
         };
         if (SwingUtilities.isEventDispatchThread()) {
            runMe.run();
         } else {
            SwingUtilities.invokeLater(runMe);
         }
      }

      public void println() {
         Runnable runMe = new Runnable() {
            public void run() {
               logArea.append("\n");
            }
         };
         if (SwingUtilities.isEventDispatchThread()) {
            runMe.run();
         } else {
            SwingUtilities.invokeLater(runMe);
         }
      }

      public void println(final String msg) {
         Runnable runMe = new Runnable() {
            public void run() {
               logArea.append(msg);
               logArea.append("\n");
            }
         };
         if (SwingUtilities.isEventDispatchThread()) {
            runMe.run();
         } else {
            SwingUtilities.invokeLater(runMe);
         }
      }
   }

}
