/*
 * Copyright (C) 2008 KenD00
 * 
 * This file is part of ConvTableView.
 * 
 * ConvTableView is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package convtableview.gui;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;

import dumphd.bdplus.ConversionTable;
import dumphd.bdplus.SubTable;

public class ConversionTableWrapper implements MutableTreeNode {

   private MutableTreeNode parent = null;
   private String nodeName = "";

   private ConversionTable table = null;
   private Vector<SubTableWrapper> tables = new Vector<SubTableWrapper>();


   public ConversionTableWrapper(ConversionTable table, String name) {
      this.table = table;
      nodeName = name;
      tables.ensureCapacity(table.size());
      Iterator<SubTable> it = table.iterator();
      while (it.hasNext()) {
         SubTableWrapper subTable = new SubTableWrapper(it.next(), this);
         tables.add(subTable);
      }
   }

   public String toString() {
      return nodeName;
   }

   public ConversionTable getConversionTable() {
      return table;
   }

   // ################################################################################################

   // ##########################
   // TreeNode Interface Methods
   // ##########################

   public Enumeration<SubTableWrapper> children() {
      return tables.elements();
   }

   public boolean getAllowsChildren() {
      return true;
   }

   public TreeNode getChildAt(int childIndex) {
      return tables.get(childIndex);
   }

   public int getChildCount() {
      return tables.size();
   }

   public int getIndex(TreeNode node) {
      return tables.indexOf(node);
   }

   public TreeNode getParent() {
      return parent;
   }

   public boolean isLeaf() {
      return (tables.size() == 0);
   }

   // ################################################################################################

   // #################################
   // MutableTreeNode Interface Methods 
   // #################################

   public void insert(MutableTreeNode child, int index) {
      // Do nothing
   }

   public void remove(int index) {
      // Do nothing
   }

   public void remove(MutableTreeNode node) {
      // Do Nothing
   }

   public void removeFromParent() {
      parent = null;
   }

   public void setParent(MutableTreeNode newParent) {
      parent = newParent;
   }

   public void setUserObject(Object object) {
      // Do nothing
   }

   // ################################################################################################

}
