/*
 * Copyright (C) 2008 KenD00
 * 
 * This file is part of ConvTableView.
 * 
 * ConvTableView is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package convtableview.gui;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;
import javax.swing.tree.TreeNode;

import dumphd.bdplus.Segment;
import dumphd.bdplus.SubTable;


public class SubTableWrapper implements TreeNode {

   private TreeNode parent = null;
   private String nodeName = "";

   private SubTable subTable = null;
   private Vector<SegmentWrapper> segments = new Vector<SegmentWrapper>();


   public SubTableWrapper(SubTable subTable, TreeNode parent) {
      this.parent = parent;
      this.subTable = subTable;
      segments.ensureCapacity(subTable.size());
      Iterator<Segment> it = subTable.iterator();
      boolean isEmpty = true;
      while (it.hasNext()) {
         SegmentWrapper segment = new SegmentWrapper(it.next(), this);
         segments.add(segment);
         if (segment.getSegment().size() > 0) {
            if (isEmpty) {
               isEmpty = false;
            }
         }
      }
      String namePrefix = null;
      if (isEmpty) {
         namePrefix = "(empty) ";
      } else {
         namePrefix = "";
      }
      nodeName = String.format("0x%1$08X: %2$sTableID %3$05d", subTable.getTableOffset(), namePrefix, subTable.getTableId());
   }

   public String toString() {
      return nodeName;
   }

   public SubTable getSubTable() {
      return subTable;
   }


   // ################################################################################################

   // ##########################
   // TreeNode Interface Methods
   // ##########################

   public Enumeration<SegmentWrapper> children() {
      return segments.elements();
   }

   public boolean getAllowsChildren() {
      return true;
   }

   public TreeNode getChildAt(int childIndex) {
      return segments.get(childIndex);
   }

   public int getChildCount() {
      return segments.size();
   }

   public int getIndex(TreeNode node) {
      return segments.indexOf(node);
   }

   public TreeNode getParent() {
      return parent;
   }

   public boolean isLeaf() {
      return (segments.size() == 0);
   }

   // ################################################################################################

}
