/*
DumpVID 0.4
2008-12-04
By KenD00

This "quick hack" is based on the CDToImg v1.01 by Truman.
I basically removed the not needed stuff and created the required CDB's like the author of that program.

The program is written and compiled under MS Visual Studio C++ 2008 Express.

Version History
---------------
0.4: 2008-12-04
- Now works for Blu-Ray too
- Detects the present media type
- Hammer mode is the default now
- Hammering is done by default with AGID 0 only now
- The Volume ID is printed in the console instead written to file
- Additional command line switches
- Code cleanup

0.31: 2007-02-23
- Fixed the bug that two bytes of the cds get lost
- Initializing the buffer with 0xFF when reading the cds to be sure that data is returned

0.3:
- Added hammering mode to grab the VolumeID when a player authenticates the drive
- Fixed BCA reading, two bytes too much were written

0.2:
- Initial public release


----------------

The following text was present in the original file, i left it in as reference.

******************************************************************************************************

CDToImg v1.01.
22 Oct 2006.
Written by Truman.

Language and type: MS Visual Studio .NET 2002, Visual C++ v7, mixed C and C++,
Application type : Win32 console.

A program that reads an entire CD-ROM (mode 1 or mode 2 form 1) 2048 bytes per
sector and writes to a file - this would be an .ISO file. The CD must be CD-ROM
mode 1 or mode 2 and non-multisession. Note, it does not check if the CD is
valid.  TAO writing mode creates link blocks and this program does not detect
them so if encountered will be considered as a normal sector or unreadable.
Unreadable sectors are not supported and will terminate the process. Only up to
80 minutes reading is supported for the moment. Note that some drives will
ignore certain read speeds, e.g. my Plextor 755A can only be set to a minimum
of 4x read speed. On some drives, reading the last few sectors or ejecting the
disc when reading in 2048 byte/sector mode results in a hang (e.g.: LG CDROM
drive: HL-DT-ST CDROM GCR-8485B 1.05).

As always I do not take any responsibilities if this tool destroys your drive or
even anything else.

This code uses:
- Win32 IOCTL function with SCSI_PASS_THROUGH_DIRECT.
- The SCSI codes used in this source were taken from the draft documents MMC1,
SPC1 and SAM1.
- MMC1 Read CD command (0xBE, CDB 12) to read sectors (2048 user bytes mode).
- Determine errors, retrieve and decode a few sense data.

Normally you need the ntddscsi.h file from Microsoft DDK CD, but I shouldn't
distribute it, so instead I have written my own my_ntddscsi.h.

If you don't have windows.h some of the define constants are listed as comments.

This is a Win32 console program and only runs in a DOS prompt under Windows
NT4/2K/XP/2003 with appropriate user rights, i.e. you need to log in as
administrator.

Read readme.txt for more info.
*/

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <winsock.h>
#include <conio.h>
#include "my_ntddscsi.h"
#include "sam.h"

//Global variables..
T_SPDT_SBUF sptd_sb;  //Includes sense buffer
unsigned char *data_buf;  //Buffer for holding transfer data from or to drive.

/*
1. Check the drive type to see if it's identified as a CDROM drive

2. Uses Win32 CreateFile function to get a handle to a drive
that is specified by cDriveLetter.

If you don't have windows.h, some of the define constants are
listed as comments.
*/
HANDLE open_volume(char drive_letter)
{
	HANDLE hVolume;
	UINT uDriveType;
	char szVolumeName[8];
	char szRootName[5];

	//Make drive_letter as a null terminated string in the form of e.g.: d:\.
	szRootName[0]=drive_letter;
	szRootName[1]=':';
	szRootName[2]='\\';
	szRootName[3]='\0';

	uDriveType = GetDriveType(szRootName);

	//#define GENERIC_READ 0x80000000L
	//#define GENERIC_WRITE 0x40000000L
	switch(uDriveType)
	{
	case DRIVE_CDROM:
		{
			printf("Drive type is recognised as CDROM/DVD.\n");
			break;
		}
	default:
		{
			printf("Drive type is not CDROM/DVD, aborting.\n");
			return INVALID_HANDLE_VALUE;  //#define INVALID_HANDLE_VALUE (long int *)-1
		}
	}

	//Make drive_letter as a null terminated string in the form of e.g.: \\.\d:.
	szVolumeName[0]='\\';
	szVolumeName[1]='\\';
	szVolumeName[2]='.';
	szVolumeName[3]='\\';
	szVolumeName[4]=drive_letter;
	szVolumeName[5]=':';
	szVolumeName[6]='\0';

	//#define FILE_SHARE_READ 0x00000001
	//#define FILE_SHARE_WRITE 0x00000002
	//#define OPEN_EXISTING 3
	//#define FILE_ATTRIBUTE_NORMAL 0x00000080
	//This will only work for CD/DVD device on W2K and higher.
	hVolume = CreateFile(szVolumeName,
		GENERIC_READ | GENERIC_WRITE,
		FILE_SHARE_READ | FILE_SHARE_WRITE,
		NULL,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		NULL);

	if(hVolume == INVALID_HANDLE_VALUE)
	{
		//Try again for NT4..
		hVolume = CreateFile(szVolumeName,
			GENERIC_READ,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL);

		if(hVolume == INVALID_HANDLE_VALUE)
		{
			printf("Could not create handle for CD/DVD device.\n");
		}
	}

	return hVolume;
}

/*
 * Closes the given handle, this function exists just as counter part to open_volume
 */
void close_volume(HANDLE hVolume) {
	CloseHandle(hVolume);
}

/* Displays sense error information. */
void disp_sense()
{
	unsigned char key;
	unsigned char ASC;
	unsigned char ASCQ;

	key=sptd_sb.SenseBuf[2] & 0x0F;  //Sense key is only the lower 4 bits
	ASC=sptd_sb.SenseBuf[12];
	ASCQ=sptd_sb.SenseBuf[13];

	printf("Sense data, key:ASC:ASCQ: %02X:%02X:%02X", key, ASC, ASCQ);

	//Decode sense key:ASC:ASCQ.
	//It's a very short list - I'm just trying to show you how to decode into text.
	//You really need to look into MMC document and change this into an exhaustive list from
	//the sense error table that is found in there.
	if(key==SEN_KEY_NO_SEN)
	{
		if(ASC==0x00)
		{
			if(ASCQ==0x00)
			{
				printf(" - No additional sense info.");  //No errors
			}
		}
	}
	else
		if(key==SEN_KEY_NOT_READY)
		{
			if(ASC==0x3A)
			{
				if(ASCQ==0x00)
				{
					printf(" - Medium not present.");
				}
				else
					if(ASCQ==0x01)
					{
						printf(" - Medium not present-tray closed.");
					}
					else
						if(ASCQ==0x02)
						{
							printf(" - Medium not present-tray open.");
						}
			}
		}

		printf("\n");
}

/*
1. Set up the sptd values.
2. Set up the CDB for SPC1 test unit ready command.
3. Send the request to the drive.
*/
BOOL test_unit_ready(HANDLE hVolume)
{
	DWORD dwBytesReturned;

	sptd_sb.sptd.Length=sizeof(SCSI_PASS_THROUGH_DIRECT);
	sptd_sb.sptd.PathId=0;    //SCSI card ID will be filled in automatically.
	sptd_sb.sptd.TargetId=0;  //SCSI target ID will also be filled in.
	sptd_sb.sptd.Lun=0;       //SCSI lun ID will also be filled in.
	sptd_sb.sptd.CdbLength=6;  //CDB size.
	sptd_sb.sptd.SenseInfoLength=MAX_SENSE_LEN;  //Maximum length of sense data to retrieve.
	sptd_sb.sptd.DataIn=SCSI_IOCTL_DATA_UNSPECIFIED; //There will be no buffer data to/from drive.
	sptd_sb.sptd.DataTransferLength=0;  //Size of buffer transfer data.
	sptd_sb.sptd.TimeOutValue=108000;  //SCSI timeout value (max 108000 sec = time 30 min).
	sptd_sb.sptd.DataBuffer=(PVOID)data_buf;
	sptd_sb.sptd.SenseInfoOffset=sizeof(SCSI_PASS_THROUGH_DIRECT);

	//CDB with values for Test Unit Ready CDB6 command.
	//The values were taken from SPC1 draft paper.
	sptd_sb.sptd.Cdb[0]=0x00;  //Code for Test Unit Ready CDB6 command.
	sptd_sb.sptd.Cdb[1]=0;
	sptd_sb.sptd.Cdb[2]=0;
	sptd_sb.sptd.Cdb[3]=0;
	sptd_sb.sptd.Cdb[4]=0;
	sptd_sb.sptd.Cdb[5]=0;
	sptd_sb.sptd.Cdb[6]=0;
	sptd_sb.sptd.Cdb[7]=0;
	sptd_sb.sptd.Cdb[8]=0;
	sptd_sb.sptd.Cdb[9]=0;
	sptd_sb.sptd.Cdb[10]=0;
	sptd_sb.sptd.Cdb[11]=0;
	sptd_sb.sptd.Cdb[12]=0;
	sptd_sb.sptd.Cdb[13]=0;
	sptd_sb.sptd.Cdb[14]=0;
	sptd_sb.sptd.Cdb[15]=0;

	ZeroMemory(sptd_sb.SenseBuf, MAX_SENSE_LEN);

	//Send the command to drive.
	return DeviceIoControl(hVolume,
		IOCTL_SCSI_PASS_THROUGH_DIRECT,
		(PVOID)&sptd_sb, (DWORD)sizeof(sptd_sb),
		(PVOID)&sptd_sb, (DWORD)sizeof(sptd_sb),
		&dwBytesReturned,
		NULL);
}

/* Sends Test Unit Ready command 3 times, check for errors & display error info. */
BOOL verified_test_unit_ready3(HANDLE hVolume)
{
	unsigned char i=3;
	BOOL success;

	/*
	Before sending the required command, here we clear any pending sense info from the drive
	which may interfere by sending Test Unit Ready command at least 3 times if neccessary.
	----------------------------------------------------------------------------------------*/
	do
	{
		printf("Sending SPC1 Test Unit CDB6 command..");
		//Sends SPC1 Test Unit Ready command to drive
		success=test_unit_ready(hVolume);
		if(success)
		{
			printf("done.\n");
			if(sptd_sb.sptd.ScsiStatus==STATUS_GOOD)
			{
				printf("Returned good status.\n");
				i=1;
			}
			else
			{
				if(sptd_sb.sptd.ScsiStatus==STATUS_CHKCOND)
				{
					disp_sense();
				}
				else
				{
					printf("Command sent but returned with an unhandled status code: %02X\n", sptd_sb.sptd.ScsiStatus);
				}

				success=FALSE;
			}
		}
		else
		{
			printf("failed.\n");
			printf("DeviceIOControl returned with failed status.\n");
		}
		i--;
	}while(i>0);

	return success;
}

/*
 * Reads the specified Feature Descriptor
 */
BOOL read_fd(HANDLE hVolume, unsigned short fd) {

	DWORD dwBytesReturned;

	sptd_sb.sptd.Length = sizeof(SCSI_PASS_THROUGH_DIRECT);
	sptd_sb.sptd.PathId = 0;    //SCSI card ID will be filled in automatically.
	sptd_sb.sptd.TargetId = 0;  //SCSI target ID will also be filled in.
	sptd_sb.sptd.Lun = 0;       //SCSI lun ID will also be filled in.
	sptd_sb.sptd.CdbLength = 9;  //CDB size.
	sptd_sb.sptd.SenseInfoLength = MAX_SENSE_LEN;  //Maximum length of sense data to retrieve.
	sptd_sb.sptd.DataIn = SCSI_IOCTL_DATA_IN;  //There will be data coming from the drive.
	sptd_sb.sptd.DataTransferLength = 2048;  //Size of data returned, 2048 bytes should be enough
	sptd_sb.sptd.TimeOutValue = 5;  //SCSI timeout value (max 108000 sec = time 30 min).
	sptd_sb.sptd.DataBuffer = (PVOID)data_buf;
	sptd_sb.sptd.SenseInfoOffset = sizeof(SCSI_PASS_THROUGH_DIRECT);

	// CDB with values for GET CONFIGURATION
	// Code for GET CONFIGURATION
	sptd_sb.sptd.Cdb[0] = 0x46;
	// 10b to deliver the requested FD
	sptd_sb.sptd.Cdb[1] = 0x02;
	// The requested FD
	sptd_sb.sptd.Cdb[2] = (fd >> 8) & 0xFF;
	sptd_sb.sptd.Cdb[3] = fd & 0xFF;
	// Reserved values
	sptd_sb.sptd.Cdb[4] = 0;
	sptd_sb.sptd.Cdb[5] = 0;
	sptd_sb.sptd.Cdb[6] = 0;
	// Allocation length for the return buffer, 2048 bytes should be enough
	sptd_sb.sptd.Cdb[7] = 0x08;
	sptd_sb.sptd.Cdb[8] = 0;
	// Control is 0, don't know whats this for
	sptd_sb.sptd.Cdb[9] = 0;
	// Zero the rest of the structure
	sptd_sb.sptd.Cdb[10] = 0;
	sptd_sb.sptd.Cdb[11] = 0;
	sptd_sb.sptd.Cdb[12] = 0;
	sptd_sb.sptd.Cdb[13] = 0;
	sptd_sb.sptd.Cdb[14] = 0;
	sptd_sb.sptd.Cdb[15] = 0;

	ZeroMemory(data_buf, 2048);
	ZeroMemory(sptd_sb.SenseBuf, MAX_SENSE_LEN);

	//Send the command to drive.
	return DeviceIoControl(hVolume,
		IOCTL_SCSI_PASS_THROUGH_DIRECT,
		(PVOID)&sptd_sb, (DWORD)sizeof(sptd_sb),
		(PVOID)&sptd_sb, (DWORD)sizeof(sptd_sb),
		&dwBytesReturned,
		NULL);
}

/*
 * Sends a READ DISC STRUCTURE command with Format Code BCA Information to the drive
 */
BOOL read_bca(HANDLE hVolume) {

	DWORD dwBytesReturned;

	sptd_sb.sptd.Length = sizeof(SCSI_PASS_THROUGH_DIRECT);
	sptd_sb.sptd.PathId = 0;    //SCSI card ID will be filled in automatically.
	sptd_sb.sptd.TargetId = 0;  //SCSI target ID will also be filled in.
	sptd_sb.sptd.Lun = 0;       //SCSI lun ID will also be filled in.
	sptd_sb.sptd.CdbLength = 12;  //CDB size.
	sptd_sb.sptd.SenseInfoLength = MAX_SENSE_LEN;  //Maximum length of sense data to retrieve.
	sptd_sb.sptd.DataIn = SCSI_IOCTL_DATA_IN;  //There will be data coming from the drive.
	sptd_sb.sptd.DataTransferLength = 2048;  //Size of read data.
	sptd_sb.sptd.TimeOutValue = 108000;  //SCSI timeout value (max 108000 sec = time 30 min).
	sptd_sb.sptd.DataBuffer = (PVOID)data_buf;
	sptd_sb.sptd.SenseInfoOffset = sizeof(SCSI_PASS_THROUGH_DIRECT);

	//CDB with values for READ DISC STRUCTURE, Format code for BCA
	//Code for READ DISC STRUCTURE
	sptd_sb.sptd.Cdb[0] = 0xAD;
	// Media Type for DVD and HD DVD is 0000b
	sptd_sb.sptd.Cdb[1] = 0;
	// Adress is reserved, so set it to 0
	sptd_sb.sptd.Cdb[2] = 0;
	sptd_sb.sptd.Cdb[3] = 0;
	sptd_sb.sptd.Cdb[4] = 0;
	sptd_sb.sptd.Cdb[5] = 0;
	// Layer is reserved, so set it to 0
	sptd_sb.sptd.Cdb[6] = 0;
	// Format for BCA: 0x03
	sptd_sb.sptd.Cdb[7] = 0x03;
	// Allocation length for the return buffer, let it be 2048
	sptd_sb.sptd.Cdb[8] = 0x08;
	sptd_sb.sptd.Cdb[9] = 0;
	// No AGID
	sptd_sb.sptd.Cdb[10] = 0;
	// Control is 0, don't know whats this for
	sptd_sb.sptd.Cdb[11] = 0;
	// Zero the rest of the structure
	sptd_sb.sptd.Cdb[12] = 0;
	sptd_sb.sptd.Cdb[13] = 0;
	sptd_sb.sptd.Cdb[14] = 0;
	sptd_sb.sptd.Cdb[15] = 0;

	ZeroMemory(data_buf, 2048);
	ZeroMemory(sptd_sb.SenseBuf, MAX_SENSE_LEN);

	//Send the command to drive.
	return DeviceIoControl(hVolume,
		IOCTL_SCSI_PASS_THROUGH_DIRECT,
		(PVOID)&sptd_sb, (DWORD)sizeof(sptd_sb),
		(PVOID)&sptd_sb, (DWORD)sizeof(sptd_sb),
		&dwBytesReturned,
		NULL);
}

/*
 * Sends a READ DISC STRUCTURE command with Format Code for Copyright Data Section to the drive.
 * The parameter sector is the starting sector to read from, values 0 to 31 are valid, at most
 * 63488 bytes ( == 31 sectors) are read.
 */
BOOL read_cds(HANDLE hVolume, unsigned char sector) {

	DWORD dwBytesReturned;

	sptd_sb.sptd.Length = sizeof(SCSI_PASS_THROUGH_DIRECT);
	sptd_sb.sptd.PathId = 0;    //SCSI card ID will be filled in automatically.
	sptd_sb.sptd.TargetId = 0;  //SCSI target ID will also be filled in.
	sptd_sb.sptd.Lun = 0;       //SCSI lun ID will also be filled in.
	sptd_sb.sptd.CdbLength = 12;  //CDB size.
	sptd_sb.sptd.SenseInfoLength = MAX_SENSE_LEN;  //Maximum length of sense data to retrieve.
	sptd_sb.sptd.DataIn = SCSI_IOCTL_DATA_IN;  //There will be data coming from the drive.
	sptd_sb.sptd.DataTransferLength = 63492;  //Max size of data that can be returned
	sptd_sb.sptd.TimeOutValue = 108000;  //SCSI timeout value (max 108000 sec = time 30 min).
	sptd_sb.sptd.DataBuffer = (PVOID)data_buf;
	sptd_sb.sptd.SenseInfoOffset = sizeof(SCSI_PASS_THROUGH_DIRECT);

	//CDB with values for READ DISC STRUCTURE, Format code for Copyright Data Section
	//Code for READ DISC STRUCTURE
	sptd_sb.sptd.Cdb[0] = 0xAD;
	// Media Type for DVD and HD DVD is 0000b
	sptd_sb.sptd.Cdb[1] = 0;
	// Adress is the starting sector to read
	sptd_sb.sptd.Cdb[2] = 0;
	sptd_sb.sptd.Cdb[3] = 0;
	sptd_sb.sptd.Cdb[4] = 0;
	sptd_sb.sptd.Cdb[5] = sector;
	// Layer is reserved, so set it to 0
	sptd_sb.sptd.Cdb[6] = 0;
	// Format for BCA: 0x03
	sptd_sb.sptd.Cdb[7] = 0x15;
	// Allocation length for the return buffer, the maximum is 63488
	sptd_sb.sptd.Cdb[8] = 0xF8;
	sptd_sb.sptd.Cdb[9] = 0x04;
	// No AGID
	sptd_sb.sptd.Cdb[10] = 0;
	// Control is 0, don't know whats this for
	sptd_sb.sptd.Cdb[11] = 0;
	// Zero the rest of the structure
	sptd_sb.sptd.Cdb[12] = 0;
	sptd_sb.sptd.Cdb[13] = 0;
	sptd_sb.sptd.Cdb[14] = 0;
	sptd_sb.sptd.Cdb[15] = 0;

	// Just to be sure that there is something returned!
	//ZeroMemory(data_buf, 63492);
	memset(data_buf, 0xFF, 63492);
	ZeroMemory(sptd_sb.SenseBuf, MAX_SENSE_LEN);

	//Send the command to drive.
	return DeviceIoControl(hVolume,
		IOCTL_SCSI_PASS_THROUGH_DIRECT,
		(PVOID)&sptd_sb, (DWORD)sizeof(sptd_sb),
		(PVOID)&sptd_sb, (DWORD)sizeof(sptd_sb),
		&dwBytesReturned,
		NULL);
}

/*
 * Sends a READ DISC STRUCTURE command with Format Code for VolumeID to the drive.
 * This command requires a succesful AACS-Authenticated drive / host configuration,
 * use it when a player authenticates the drive to drive by.
 * agid is the AGID of the authenticated connection to use to request the VolumeID.
 * This only works if the bus traffic is not encrypted.
 */
BOOL read_vid(HANDLE hVolume, unsigned char mediaType, unsigned char agid) {

	DWORD dwBytesReturned;

	sptd_sb.sptd.Length = sizeof(SCSI_PASS_THROUGH_DIRECT);
	sptd_sb.sptd.PathId = 0;    //SCSI card ID will be filled in automatically.
	sptd_sb.sptd.TargetId = 0;  //SCSI target ID will also be filled in.
	sptd_sb.sptd.Lun = 0;       //SCSI lun ID will also be filled in.
	sptd_sb.sptd.CdbLength = 12;  //CDB size.
	sptd_sb.sptd.SenseInfoLength = MAX_SENSE_LEN;  //Maximum length of sense data to retrieve.
	sptd_sb.sptd.DataIn = SCSI_IOCTL_DATA_IN;  //There will be data coming from the drive.
	sptd_sb.sptd.DataTransferLength = 2048;  //Size of data returned, 2048 bytes should be enough
	sptd_sb.sptd.TimeOutValue = 5;  //SCSI timeout value (max 108000 sec = time 30 min).
	sptd_sb.sptd.DataBuffer = (PVOID)data_buf;
	sptd_sb.sptd.SenseInfoOffset = sizeof(SCSI_PASS_THROUGH_DIRECT);

	//CDB with values for READ DISC STRUCTURE, Format code for VolumeID
	//Code for READ DISC STRUCTURE
	sptd_sb.sptd.Cdb[0] = 0xAD;
	// Media Type for DVD and HD DVD is 0000b
	sptd_sb.sptd.Cdb[1] = mediaType;
	// Adress is reserved
	sptd_sb.sptd.Cdb[2] = 0;
	sptd_sb.sptd.Cdb[3] = 0;
	sptd_sb.sptd.Cdb[4] = 0;
	sptd_sb.sptd.Cdb[5] = 0;
	// Layer is reserved, so set it to 0
	sptd_sb.sptd.Cdb[6] = 0;
	// Format for VolumeID: 0x80
	sptd_sb.sptd.Cdb[7] = 0x80;
	// Allocation length for the return buffer, 2048 bytes should be enough
	sptd_sb.sptd.Cdb[8] = 0x08;
	sptd_sb.sptd.Cdb[9] = 0;
	// AGID to use, AGID is at bit 6 and 7
	sptd_sb.sptd.Cdb[10] = agid << 6;
	// Control is 0, don't know whats this for
	sptd_sb.sptd.Cdb[11] = 0;
	// Zero the rest of the structure
	sptd_sb.sptd.Cdb[12] = 0;
	sptd_sb.sptd.Cdb[13] = 0;
	sptd_sb.sptd.Cdb[14] = 0;
	sptd_sb.sptd.Cdb[15] = 0;

	ZeroMemory(data_buf, 2048);
	ZeroMemory(sptd_sb.SenseBuf, MAX_SENSE_LEN);

	//Send the command to drive.
	return DeviceIoControl(hVolume,
		IOCTL_SCSI_PASS_THROUGH_DIRECT,
		(PVOID)&sptd_sb, (DWORD)sizeof(sptd_sb),
		(PVOID)&sptd_sb, (DWORD)sizeof(sptd_sb),
		&dwBytesReturned,
		NULL);
}

/*
 * Main function to detect the media type.
 * Sets mediaType to 2 if an error ocurred or the media is not of type HD-DVD or Blu-Ray.
 */
BOOL detect_mediaType(HANDLE hVolume, unsigned char *mediaType) {
	BOOL success;
	// Retrieve the Read Blu-Ray Feature Descriptor first
	unsigned short fd = 0x40;

	// Set mediaType to 2 to be sure that the check if Blu-Ray should be testet doesn't fail
	*mediaType = 2;
	while (true) {
		// Retrieve the Feature Descriptor
		success = read_fd(hVolume, fd);
		if (success) {
			if (sptd_sb.sptd.ScsiStatus == STATUS_GOOD) {
				// If the data length is > 4 the Feature Descriptor was retrieved
				if (ntohl(*((unsigned int *)data_buf)) > 4) {
					// Check if it is active
					if ((*(data_buf + 10) & 0x1) == 1) {
						// Active, signal proper media type
						if (fd == 0x40) {
							*mediaType = 1;
						} else {
							*mediaType = 0;
						}
						break;
					}
				}
				// Check if should be tested for another Feature Descriptor
				if (*mediaType == 2 && fd == 0x40) {
					// Retrieve Read HD-DVD Feature Descriptor
					fd = 0x50;
				} else {
					break;
				}
			} else {
				if (sptd_sb.sptd.ScsiStatus == STATUS_CHKCOND) {
					disp_sense();
				} else {
					printf("Command sent but returned with an unhandled status code: %02X\n", sptd_sb.sptd.ScsiStatus);
				}
				printf("Aborting process.\n");
				success = FALSE;
				break;
			} 
		} else {
			printf("Read Feature Descriptor command failed!\n");
			printf("Aborting process.\n");
			break;
		}
	}

	return success;
}

/*
 * Main function, reads the BCA and Copyright Data Section and writes them to file.
 */
BOOL dump_vid(HANDLE hVolume) {
	BOOL success;
	FILE *bca_ptr;
	FILE *cds_ptr;

	printf("Reading BCA...\n");
	success = read_bca(hVolume);
	if (success) {
		if (sptd_sb.sptd.ScsiStatus == STATUS_GOOD) {
			bca_ptr = fopen("bca.bin", "wb");
			if (bca_ptr != NULL) {
				// The actual BCA data starts at byte 4, the first two bytes are the length of the returned data
				// (in Big Endian, must be converted to the host byte order), the next two bytes are reserved and
				// do not belong to the BCA, but they are included in the length field so they must be subtracted
				fwrite(data_buf + 4, ntohs(*((unsigned short *)data_buf)) - 2, 1, bca_ptr);
				if (ferror(bca_ptr)) {
					printf("Write file error!\n");
					printf("Aborting process.\n");
					success = FALSE;
				}
				fclose(bca_ptr);
			} else {
				printf("Could not create file!\n");
				printf("Aborting process.\n");
				success = FALSE;
			}
		} else {
			if (sptd_sb.sptd.ScsiStatus == STATUS_CHKCOND) {
				disp_sense();
			} else {
				printf("Command sent but returned with an unhandled status code: %02X\n", sptd_sb.sptd.ScsiStatus);
			}
			printf("Aborting process.\n");
			success = FALSE;
		} 
	} else {
		printf("Read BCA command failed!\n");
		printf("Aborting process.\n");
	}
	printf("Reading Copyright Data Section...\n");
	// Read the first 31 sectors
	success = read_cds(hVolume, 0);
	if (success) {
		if (sptd_sb.sptd.ScsiStatus == STATUS_GOOD) {
			cds_ptr = fopen("cds.bin", "wb");
			if (cds_ptr != NULL) {
				// The actual data starts at byte 4, size is 63488 bytes == 31 sectors
				fwrite(data_buf + 4, 63488, 1, cds_ptr);
				if (ferror(cds_ptr)) {
					printf("Write file error!\n");
					printf("Aborting process.\n");
					success = FALSE;
				} else {
					// Read the missing last sector
					success = read_cds(hVolume, 31);
					if (success) {
						if (sptd_sb.sptd.ScsiStatus == STATUS_GOOD) {
							// The actual data starts at byte 4, the size is 2048 bytes == 1 sector
							fwrite(data_buf + 4, 2048, 1, cds_ptr);
							if (ferror(cds_ptr)) {
								printf("Write file error!\n");
								printf("Aborting process.\n");
								success = FALSE;
							}
						} else {
							if (sptd_sb.sptd.ScsiStatus == STATUS_CHKCOND) {
								disp_sense();
							} else {
								printf("Command sent but returned with an unhandled status code: %02X\n", sptd_sb.sptd.ScsiStatus);
							}
							printf("Aborting process.\n");
							success = FALSE;
						}
					} else {
						printf("Read Copyright Data Section command failed!\n");
						printf("Aborting process.\n");
					}
				}
				fclose(cds_ptr);
			} else {
				printf("Could not create file!\n");
				printf("Aborting process.\n");
				success = FALSE;
			}
		} else {
			if (sptd_sb.sptd.ScsiStatus == STATUS_CHKCOND) {
				disp_sense();
			} else {
				printf("Command sent but returned with an unhandled status code: %02X\n", sptd_sb.sptd.ScsiStatus);
			}
			printf("Aborting process.\n");
			success = FALSE;
		} 
	} else {
		printf("Read Copyright Data Section command failed!\n");
		printf("Aborting process.\n");
	}

	return success;
}

/*
 * Main function to hammer the drive to get the VolumeID. Sleep_amount is the number of milliseconds to wait between each request.
 */
BOOL hammer_vid(HANDLE hVolume, unsigned char mediaType, unsigned int sleep_amount, bool cycleAgids) {
	BOOL success;
	unsigned char agid = 0;

	// Hammer the drive until the command was successful
	printf("Press ENTER to start hammering\n");
	unsigned int dummy = 0;
	scanf("%c", &dummy);
	printf("Hammering drive with a %ums delay...\n", sleep_amount);
	while (true) {
		// Try to read the VID
		success = read_vid(hVolume, mediaType, agid);
		if (success) {
			if (sptd_sb.sptd.ScsiStatus == STATUS_GOOD) {
				// The actual data starts at byte 4, size is 16 bytes
				printf("VID: ");
				for (int i = 0; i < 16; i++) {
					printf("%02X", *(data_buf + 4 + i));
				}
				printf("\n");
				break;
			}
		} else {
			printf("Read Volume ID command failed!\n");
			printf("Aborting process.\n");
			break;
		}
		// Increment AGID, range is 0 to 3
		if (cycleAgids) {
			agid = (agid + 1) % 4;
		}
		// Sleep a bit to reduce the hammering interval
		Sleep(sleep_amount);
		// If aborted with a keystroke consume the character and signal an error
		if (_kbhit()) {
			_getch();
			success = false;
			break;
		}
	}

	return success;
}

void printUsage() {
	printf("Usage: dumpvid [options] <drive letter> [delay]\n");
	printf("\n");
	printf("Hammers the drive with Volume ID requests waiting delay ms between\n");
	printf("each attempt (100 ms if no delay is specified). Use this method while\n");
	printf("a software player authenticates the drive.\n");
	printf("Press any key to abort the hammering process.\n");
	printf("\n");
	printf("Options:\n");
	printf("-c        : Cycle AGID's during hammering (all AGID's will be used),\n");
	printf("            if this option is not specified only AGID 0 will be used\n");
	printf("-d        : Dump the BCA/CDS section instead hammering. This works only\n");
	printf("            for HD-DVDs. Creates bca.bin and cds.bin in the current directory.\n");
	printf("--force-bd: Force Blu-Ray mode\n");
	printf("--force-hd: Force HD-DVD mode\n");
	printf("\n");
	printf("Example: Hammer drive d with a 10 ms delay\n");
	printf("         dumpvid d 10\n");
	return;
}

int main(int argc, char *argv[]) {
	// Letter of the drive to use
	char driveLetter;
	// Mediatype of the disc, 0 = HD-DVD, 1 = Blu-Ray, 2 = Detect
	unsigned char mediaType = 2;
	// If true, all AGIDS will be used during hammering, otherwise only AGID 0
	bool cycleAgids = false;
	// If true, BCA/CDS will be dumped, otherwise the VID will get hammered
	bool dumpMode = false;
	// Delay in ms between hammer attempts
	unsigned int delay = 100;

	// Handle of the drive
	HANDLE hDrive;
	// Returnvalue of this function
	// 0 = OK
	// 1 = Cmdline error
	// 2 = Drive I/O error
	int rv = 1;

	printf("DumpVID 0.4 by KenD00\n\n");
	
	// Very simple command line parser
	if (argc < 2) {
		printUsage();
		return rv;
	}
	int currentArg = 1;
	// Parse options
	for (; currentArg < argc; currentArg++) {
		if (argv[currentArg][0] == '-') {
			if (strlen(argv[currentArg]) < 2) {
				printf("Unknown option: \"\"\n");
				return rv;
			}
			if (argv[currentArg][1] == '-') {
				// Long option
				if (strlen(argv[currentArg]) < 3) {
					printf("Unknown long option: \"\"\n");
					return rv;
				}
				char *lopt = argv[currentArg] + 2;
				if (strcmp(lopt, "force-hd") == 0) {
					mediaType = 0;
				} else if (strcmp(lopt, "force-bd") == 0) {
					mediaType = 1;
				} else {
					printf("Unknown long option: \"%s\"\n", lopt);
					return rv;
				}
			} else {
				// Short option
				for (unsigned int currentOpt = 1; currentOpt < strlen(argv[currentArg]); currentOpt++) {
					switch (argv[currentArg][currentOpt]) {
					case 'c':
						cycleAgids = true;
					break;
					case 'd':
						dumpMode = true;
					break;
					default:
						printf("Unknown option: \"%c\"\n", argv[currentArg][currentOpt]);
					return rv;
					}
				}
			}
		} else {
			// Not an option
			break;
		}
	}
	// Parse drive
	if (currentArg == argc) {
		printf("No drive letter specified\n");
		return rv;
	}
	driveLetter = argv[currentArg++][0];
	// Parse sleep amount
	if (currentArg < argc) {
		delay = atoi(argv[currentArg++]);
		// Don't allow 0 ms delay, this may also indicate a parsing error
		if (delay == 0) {
			printf("Invalid delay, must be a number greater than 0\n");
			return rv;
		}
	}
	// Check for remaining, invalid stuff
	if (currentArg < argc) {
		printf("Invalid parameter: \"%s\"\n", argv[currentArg]);
		return rv;
	}
	// End command line parser

	//65536 data transfer buffer.
	data_buf = (unsigned char *)malloc(65536);
	
	// Default error is now I/O error
	rv = 2;
	
	// Open the drive
	hDrive = open_volume(driveLetter);

	if (hDrive != INVALID_HANDLE_VALUE) {
		// Test if drive is ready
		printf("\n");
		if (verified_test_unit_ready3(hDrive)) {
			printf("\n");
			// Detect mediatype if necessary
			if (mediaType == 2) {
				if (detect_mediaType(hDrive, &mediaType)) {
					if (mediaType != 2) {
						printf("Detected media type: ");
						if (mediaType == 0) {
							printf("HD-DVD\n");
						} else {
							printf("Blu-Ray\n");
						}
					} else {
						printf("Could not detect media type, use a force option to select one\n");
					}
				}
			}

			// Process if media type was detected
			if (mediaType != 2) {
				printf("\n");
				// Perform desired action
				if (dumpMode) {
					if (mediaType == 0) {
						if (dump_vid(hDrive)) {
							printf("Dump finished\n");
							rv = 0;
						} else {
							printf("Dump failed from drive %c:\n", driveLetter);
						}
					} else {
						printf("Cannot dump from non HD-DVD media!\n");
						rv = 1;
					}
				} else {
					if (hammer_vid(hDrive, mediaType, delay, cycleAgids)) {
						printf("Hammering finished\n");
						rv = 0;
					} else {
						printf("Hammering failed from drive %c:\n", driveLetter);
					}
				}
			}
		} else {
			printf("\n");
			printf("Drive is not ready!\n");
			printf("Aborting process.\n");
		}

		// Finally close the drive
		close_volume(hDrive);
	}
	
	// Free the transfer buffer
	free(data_buf);

	return rv;
}
